/*                                                               -*- C -*-
 *  AAFPORT.C
 *
 *  AAF Exporter/Importer functions for Astrolog.
 *  Copyright (C) 1996 by Tobias Ferber, ferber@rpk.mach.uni-karlsruhe.de
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; either version 1 of the License,
 *  or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: aafport.c,v 1.3 1996/06/12 23:10:28 tf Exp $ */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "astrolog.h"

/* The following data tables are index compatible with those in
   the "data.c" file.  Items known to Astrolog but unknown to AAF
   are represented by an empty string "". */

static
CONST char *szSignAbbrevAAF[cSign+1] = { "",
  "AR","TA","GE","CN","LE","VI","LI","SC","SG","CP","AQ","PI"
};

static
CONST char * ARR szSystemAAF[cSystem] = {
  "PL","GO","EA","CA","ME","RE","PO","MO","TO","","EM","","",""
};

static
char *ARR szObjAbbrevAAF[objMax] = {  "", /* Earth */
  "SO","MO","ME","VE","MA","JU","SA","UR","NE","PL", /* Planets ..oMain */
  "CH","CE","PA","JO","VA", /* Asteroids */
  "?N","LL","",  "",  "",   /* Others .. oCore */
  "01","02","03","04","05","06","07","08","09","10","11","12", /* cuspLo..cuspHi */
  "CU","HA","ZE","KR","AP","AD","VU","" /* uranLo..uranHi */
  /* Stars are not available in the AAF */
};

static 
char * ARR szAspectAbbrevAAF[cAspect+1] = { "",
  "CON","OPP","SQU","TRI","SEX",
  "QCX","SSX","SSQ","SES","QUI","BQI",     
  "SQI","051","040","080","102","154","160"
};

/* Make a copy of the given string and perform some conversions
   where characters would hurt the AAF property when printed */
static char *
aafstr(str)
     char *str;
{
  char *copy_of_str= (char *)0;
  
  if(str && *str)
    {
      if( (copy_of_str= strdup(str)) )
	{
	  register char *s;

	  /* Now we traverse the string and perform some cosmetics */
	  for(s= copy_of_str; *s; s++)
	    {
	      if(*s==',')
		*s= ';';

	      else if(*s=='\n' || *s=='\r' || *s=='\t' || *s=='\v')
		*s= ' ';
	    }
	}
    }

  return copy_of_str;
}

/* Make a printable string from given zodiac angle
   similar to the SzZodiac() function in "general.c" */
static char *
SzPosAAF(deg)
     real deg;
{
  static char buf[11];

  deg= Mod(deg + (is.fSeconds ? rRound/60.0/60.0 : rRound/60.0));
  
  /* SPOS position encoding: deg:min[:sec] */
  if( us.fSidereal ) /* -s */
    {
      sprintf(buf, "%3d:%02d",
	      (int)deg,
	      (int)(RFract(deg)*60.0) );
    }

  /* TPOS position encoding: deg/sign/min[:sec] */
  else /* Tropical */
    {
      int sign= (int)(deg / 30.0);
      
      sprintf(buf, "%2d%2s%02d",
	      (int)deg - sign*30,
	      szSignAbbrevAAF[sign+1],
	      (int)(RFract(deg)*60.0) );
    }

  if ( is.fSeconds ) /* -b0 */
    sprintf(&buf[6], ":%02d",
	    (int)(RFract(deg)*60.0*60.0) );
  
  return buf;
}

/*
 *  AAF exporter function
 */

int
astrolog_export_aaf(fp)
     FILE *fp;
{
  int err= 0;

  /* Write #A93: chunk data */
  {
    char *nam= aafstr( ciMain.nam );
    char *loc= aafstr( ciMain.loc );
    
    fprintf(fp, "#A93:%s,%s,%s,%2d.%02d.%d,%2dh%02d:00,%s,%s\n",

	    /* last name, first name, sex */
	    nam ? nam : "Unnamed", "", "",
	    
	    /* date */
	    Day, Mon, Yea,
	    
	    /* time */
	    (int)NFloor(Tim),
	    (int)(RFract(RAbs(Tim))*100.0+rRound/60.0),
	    
	    /* place/location, country */
	    loc ? loc : "Nowhere", ""
	    );

    if(nam)
      free(nam);

    if(loc)
      free(loc);
  }

  /* Write #B93: chunk data */
  {
    fprintf(fp, "#B93:%012.5f,%2d%c%02d,%2d%c%02d,%2dh%c%02d,%d\n",

	    /* Julian date */
	    JulianDayFromTime(T),
	    
	    /* Latitude */	    
	    (int)RFloor(RAbs(Lat)),         
	    (Lat < 0.0) ? 'S' : 'N',
	    (int)(RFract(RAbs(Lat))*100.0+rRound),
	    
            /* Longitude */
	    (int)RFloor(RAbs(Lon)),         
	    (Lon < 0.0) ? 'E' : 'W',
	    (int)(RFract(RAbs(Lon))*100.0+rRound),

	    /* Time zone (GMT offset) */
	    (int)RAbs(Zon),
	    (Zon < 0.0) ? 'E' : 'W',
	    (int)(RFract(RAbs(Zon))*100.0+rRound/60.0),

	    /* Daylight savings time */
	    (int)Dst );
  }

  /* Write #COM: chunk (Comment) */
  fprintf(fp, "#COM:Generated by %s version %s\n", szAppName, szVersionCore);
  
  /* Write #ZNAM: chunk (Name of the time zone) */
  {
    int i;
    for(i=0; i<cZone; i++)
      if( rZon[i] == Zon )
	break;
    
    fprintf(fp, "#ZNAM:%s\n", (i < cZone) ? szZon[i] : "LMT");
  }

  /* Write #H_*POS*: chunk if the AAF knows the house system and
     the positions are either geocentric or heliocentric */
  if( *(szSystemAAF[us.nHouseSystem]) && ((us.objCenter == 1) || (us.objCenter == 0)) )
    {
      fprintf(fp, "#H_%cPOS%c:",
	      us.objCenter == 0 ? 'H' : (us.fSidereal ? 'S' : 'T'),
	      is.fSeconds ? 's' : 'm' );
      
      fprintf(fp, "%s,", SzPosAAF(planet[cuspLo + 0]));  /* AC */
      fprintf(fp, "%s,", SzPosAAF(planet[cuspLo + 9]));  /* MC */

      fprintf(fp, "%s,", SzPosAAF(chouse[ 1]));  /* 01 */
      fprintf(fp, "%s,", SzPosAAF(chouse[ 2]));  /* 02 */
      fprintf(fp, "%s,", SzPosAAF(chouse[ 3]));  /* 03 */

      fprintf(fp, "%s,", SzPosAAF(chouse[10]));  /* 10 */
      fprintf(fp, "%s,", SzPosAAF(chouse[11]));  /* 11 */
      fprintf(fp, "%s,", SzPosAAF(chouse[12]));  /* 12 */

      fprintf(fp, "%s\n", szSystemAAF[us.nHouseSystem]);
    }

  /* Write #*POS*: chunk for all unrestricted objects */
  if( (us.objCenter == 1) || (us.objCenter == 0) )  
    {
      int i,n;
      fprintf(fp, "#%cPOS%c:",
	      us.objCenter == 0 ? 'H' : (us.fSidereal ? 'S' : 'T'),
	      is.fSeconds ? 's' : 'm' );
      
      /* AAF does not know about stars i.e objecs > oNorm */
      for(i=1,n=0; i<=oNorm; i++)
	{
	  /* Skip objects which are not available in the AAF,
	     or restricted.  Ignore house cusps here as well. */
	  if( ignore[i] || !szObjAbbrevAAF[i][0] || (cuspLo <= i && i <= cuspHi) )
	    continue;
	  
	  /* Special treatment for true node vs. north node */
	  if( i == 16 ) /* Node */
	    fprintf(fp, "%s%s%s",
		    (++n>1) ? "," : "",
		    us.fTrueNode ? "TN" : "NN",  /* -Yn */
		    SzPosAAF(planet[i]) );
	  else
	    fprintf(fp, "%s%s%s",
		    (++n>1) ? "," : "",
		    szObjAbbrevAAF[i],
		    SzPosAAF(planet[i]) );
	}
      fprintf(fp,"\n");
    }
  
  /* FIXME! #ASPs chunk */
  
  return err;
}

#ifdef HAVE_LIBAAF

/*
 *  The following chunk ID specific AAF importer functions are
 *  called by the invoke_aafchunk_importer() function. They
 *  import the chunk data into Astrolog's internal data structures.
 */

#include "aaf.h"

static int
import_a93(this)
     aafchunk_t *this;
{
  int err= 0;

  if( aafchunk_strsplit(this,",",1) > 0 )
    { 
      /* name */
      if( (this->num_fields > 2) )
      {
	char *buf=
	  (char *)malloc(  strlen( (this->fields)[0] ) + 1
			 + strlen( (this->fields)[1] ) + 1 
			 + 1 + strlen( (this->fields)[2] ) + 1 + 1 );

	if( buf )
	  {
	    sprintf(buf, "%s %s (%s)",
		    (this->fields)[1],
		    (this->fields)[0],
		    (this->fields)[2] );

	    ciCore.nam= SzPersist( buf );
	    free(buf);
	  }
	else
	  err= __LINE__, PrintError("Ran out of memory!");
      }

      /* location */
      if( (this->num_fields > 6) )
      {
	char *buf=
	  (char *)malloc(  strlen( (this->fields)[5] ) + 2
			 + strlen( (this->fields)[6] ) + 1 );

	if( buf )
	  {
	    sprintf(buf,"%s, %s",
		    (this->fields)[5],
		    (this->fields)[6] );		    
	
	    ciCore.loc= SzPersist( buf );
	  }
      }

      
      /* date */
      if( (this->num_fields > 3) && *((this->fields)[3]) )      
	{
	  int d,m,y;
	  char *c;

	  err= aaf_parse_date( (this->fields)[3], &d,&m,&y, &c );

	  if( err )
	    PrintWarning("Error parsing date in #A93 chunk.");

	  /* Make sure we have a date in the Gregorian calendar */
	  else if( !c || *c=='\0' || *c=='g' || *c=='G' )
	    {
	      if( !FValidMon(m) ) 
		err= __LINE__,
		PrintWarning("Month value out of range.");
	      
	      if( !FValidYea(y) )		
		err= __LINE__,
		PrintWarning("Year value out of range.");

	      if( !FValidDay(d,m,y) ) 
		err= __LINE__,
		PrintWarning("Day value out of range.");
	      
	      if( !err ) 
		DD=d, MM=m, YY=y;
	    }
	  else
	    err= __LINE__,	      
	    PrintWarning("Calendar type of date in #A93 chunk not supported.");
	}
      
      /* time */
      if( (this->num_fields > 4) && *((this->fields)[4]) )      
	{
	  int h,m,s;
	  char *x, buf[20];

	  err= aaf_parse_time( (this->fields)[4], &h,&m,&s, &x );

	  sprintf(buf,"%d:%02d", h,m);
	  TT= RParseSz(buf, pmTim);
	}
    }
  else
    {
      PrintError("Error splitting #A93 chunk into fields");
      err= __LINE__;
    }

  return err;
}

static int
import_b93(this)
     aafchunk_t *this;
{
  int err= 0;

  if( aafchunk_strsplit(this,",",1) > 0 )
    {
      /* julian date */
      if( (this->num_fields > 0) && *((this->fields)[0]) )
	{
	  /* FIXME */
	}

      /* latitude */
      if( (this->num_fields > 1) && *((this->fields)[1]) )
	{
	  AA= RParseSz( (this->fields)[1], pmLat );
	}
      
      /* longitude */
      if( (this->num_fields > 2) && *((this->fields)[2]) )
	{
	  OO= RParseSz( (this->fields)[2], pmLon );	  
	}

      /* time zone */
      if( (this->num_fields > 3) && *((this->fields)[3]) )
	{
	  /*ZZ= -RParseSz( (this->fields)[3], pmZon );*/

	  int h,m,s;
	  char *x, buf[20];
	  
	  aaf_parse_time( (this->fields)[3], &h,&m,&s, &x );

	  ZZ= (real)h + (real)m / 100.0;  /* <== not 60.0 */

	  if( x && (*x=='E' || *x=='e') )
	    ZZ= -ZZ;
	}

      /* daylight savings time */
      if( (this->num_fields > 4) && *((this->fields)[4]) )
	{
	  SS= RParseSz( (this->fields)[4], pmDst );
	}
    }
  else
    {
      PrintError("Error splitting #B93 chunk into fields");
      err= __LINE__;
    }

  return err;
}



/*
 *  The import handler looks up the chunk ID in the following
 *  data table and invokes the chunk importer function.
 */

struct { char *id; int (*funptr)(); } import_handlers[] = {
  { "#A93",  import_a93 },
  { "#B93",  import_b93 },
  { NULL, NULL }
};

static int
invoke_aafchunk_importer(this)
     aafchunk_t *this;
{
  int err= -1;

  int h;  
  size_t this_len= strlen( this->id );

  for(h=0; (import_handlers[h]).id; h++)
    {
      char *id= (import_handlers[h]).id;
      size_t id_len= strlen( id );

      if( this_len >= id_len )
	{
	  size_t n= this_len - id_len;
	  
	  if( stricmp(&(this->id[n]), id) == 0 )
	    {
	      err= (import_handlers[h]).funptr ?
		(import_handlers[h]).funptr(this) : -1;

	      break;
	    }
	}
    }
  
  return err;
}

/*
 *  Main importer interface function
 */

int
astrolog_import_aaf(fname)
     char *fname;
{
  int err= 0;

  FILE *fp= aaf_open(fname);

  if( fp )
    {
      /* Read the next #A93 chunk from the given stream */
      aafchunk_t *a93= aafchunk_read_until( fp, "#A93", (aafchunk_t *)0 );
      
      /* Read all subsequent children of this chunk and forget
	 about the next #A93 chunk (if any) */
      
      if( a93 )
	{
	  aafchunk_dispose( aafchunk_read_until(fp,"#A93",a93) );
	  err= aafchunk_fold(a93);
	  
	  if( !err )
	    {
	      aafchunk_t *chunk;

	      /* We don't have persistent parameter strings here
	         since we'll dispose our chunks after we read them. */
	      _bool old_fSzPersist= is.fSzPersist;
	      is.fSzPersist= fFalse;
      
	      /* Invoke the importer for the #A93 chunk ... */
	      err= invoke_aafchunk_importer(a93);
	      
	      /* ... and for all of it's children */
	      for(chunk=a93->child; chunk && (err < 1); chunk= chunk->succ)
		err= invoke_aafchunk_importer(chunk);
	      
	      if(err > 0)
		PrintError("AAF chunk importer failed");
	      else
		err= 0;

	      /* restore the old argument persistence indicator */
	      is.fSzPersist= old_fSzPersist;
	    }
	  else
	    PrintError("Error folding AAF record");
	  
	  /* Free the whole record w/ all it's children */
	  a93= aafchunk_dispose(a93);
	}
      else
	{
	  err= __LINE__;
	  PrintError("No #A93 chunk found.");	  
	}

      aaf_close(fp);
    }
  else
    {
      err= __LINE__;
      PrintError("Could not access given AAF database.");
    }
  
  return err;
}

#endif /* HAVE_LIBAAF */
